#include <QTextStream>
#include <QHeaderView>
#include <QTreeWidgetItem>
#include <QStringList>
#include <QFile>
#include <QFontMetrics>
#include <QFont>
#include <QTimer>
#include <QMap>
#include <QSet>
#include <QBitArray>

#include "gamelist.h"
#include "qmc2main.h"
#include "procmgr.h"
#include "options.h"
#include "preview.h"
#include "flyer.h"

// external global variables
extern MainWindow *qmc2MainWindow;
extern ProcessManager *qmc2ProcessManager;
extern Options *qmc2Options;
extern QSettings *qmc2Config;
extern bool qmc2ReloadActive;
extern bool qmc2EarlyReloadActive;
extern bool qmc2StopParser;
extern bool qmc2StartingUp;
extern bool qmc2VerifyActive;
extern bool qmc2FilterActive;
extern int qmc2GamelistResponsiveness;
extern Preview *qmc2Preview;
extern Flyer *qmc2Flyer;
extern QTreeWidgetItem *qmc2CurrentItem;
extern QMap<QString, QTreeWidgetItem *> qmc2GamelistItemMap;
extern QMap<QString, QString> qmc2GamelistNameMap;
extern QMap<QString, QString> qmc2GamelistStatusMap;
extern int qmc2SortCriteria;
extern Qt::SortOrder qmc2SortOrder;
extern QBitArray qmc2Filter;

// local variables
QStringList Gamelist::phraseTranslatorList;
int numVerifyRoms = 0;
QString verifyLastLine;
#ifdef SDLMAME
QStringList verifiedList;
#endif

Gamelist::Gamelist(QObject *parent)
  : QObject(parent)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::Gamelist()");
#endif

  numGames = numTotalGames = numCorrectGames = numMostlyCorrectGames = numIncorrectGames = numUnknownGames = numNotFoundGames = numSearchGames = -1;
  loadProc = verifyProc = NULL;

  QString imgDir = qmc2Config->value("Frontend/FilesAndDirectories/DataDirectory", "data/").toString() + "img/";
  qmc2SmallGhostImageIcon.addFile(imgDir + "ghost_small.png");
  qmc2UnknownImageIcon.addFile(imgDir + "sphere_blue.png");
  qmc2CorrectImageIcon.addFile(imgDir + "sphere_green.png");
  qmc2MostlyCorrectImageIcon.addFile(imgDir + "sphere_yellowgreen.png");
  qmc2IncorrectImageIcon.addFile(imgDir + "sphere_red.png");
  qmc2NotFoundImageIcon.addFile(imgDir + "sphere_grey.png");

  if ( phraseTranslatorList.count() == 0 )
    phraseTranslatorList << tr("good") << tr("bad") << tr("preliminary") << tr("supported") << tr("unsupported")
                         << tr("imperfect") << tr("yes") << tr("no") << tr("baddump") << tr("nodump")
                         << tr("vertical") << tr("horizontal") << tr("raster") << tr("Unknown")
                         << tr("On") << tr("Off") << tr("audio") << tr("Unused");
}

Gamelist::~Gamelist()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::~Gamelist()");
#endif

  if ( loadProc )
    loadProc->terminate();

  if ( verifyProc )
    verifyProc->terminate();
}

void Gamelist::enableWidgets(bool enable)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::enableWidgets(bool enable = " + QString(enable ? "TRUE" : "FALSE") + ")");
#endif

  qmc2Options->toolButtonBrowseFont->setEnabled(enable);
  qmc2Options->toolButtonBrowseTemporaryFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseFrontendLogFile->setEnabled(enable);
  qmc2Options->toolButtonBrowsePreviewDirectory->setEnabled(enable);
  qmc2Options->toolButtonBrowseDataDirectory->setEnabled(enable);
  qmc2Options->toolButtonBrowseExecutableFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseMAMELogFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseOptionsTemplateFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseFavoritesFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseHistoryFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseROMStateCacheFile->setEnabled(enable);
  qmc2Options->toolButtonBrowseFlyerDirectory->setEnabled(enable);
  qmc2Options->toolButtonShowC->setEnabled(enable);
  qmc2Options->toolButtonShowM->setEnabled(enable);
  qmc2Options->toolButtonShowI->setEnabled(enable);
  qmc2Options->toolButtonShowN->setEnabled(enable);
  qmc2Options->toolButtonShowU->setEnabled(enable);

  qmc2Options->comboBoxSortCriteria->setEnabled(enable);
  qmc2Options->comboBoxSortOrder->setEnabled(enable);
}

void Gamelist::load()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::load()");
#endif

  qmc2ReloadActive = qmc2EarlyReloadActive = TRUE;
  qmc2StopParser = FALSE;
  qmc2GamelistItemMap.clear();
  qmc2GamelistNameMap.clear();
  qmc2GamelistStatusMap.clear();

  enableWidgets(FALSE);

  numGames = numTotalGames = numCorrectGames = numMostlyCorrectGames = numIncorrectGames = numUnknownGames = numNotFoundGames = numSearchGames = -1;
  qmc2MainWindow->treeWidgetGamelist->clear();
  qmc2MainWindow->listWidgetSearch->clear();
  qmc2CurrentItem = NULL;
  qmc2Preview->repaint();
  qmc2Flyer->repaint();
  QTreeWidgetItem *dummyItem = new QTreeWidgetItem(qmc2MainWindow->treeWidgetGamelist);
  dummyItem->setText(0, tr("Waiting for data..."));
  qmc2MainWindow->labelGamelistStatus->setText(status());

  // determine number of supported games
  qmc2MainWindow->log(LOG_QMC2, tr("determining total number of supported games"));

  QStringList args;
  QString command;
#ifdef SDLMAME
  QProcess commandProc;
  commandProc.setStandardOutputFile(qmc2Config->value("Frontend/FilesAndDirectories/TemporaryFile").toString());
  commandProc.setStandardErrorFile("/dev/null");
  args << "-listfull";
#else
  command = qmc2Config->value("MAME/FilesAndDirectories/ExecutableFile").toString() + " -l -nolcf -err /dev/null -out " + qmc2Config->value("Frontend/FilesAndDirectories/TemporaryFile").toString();
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: executing command: command = \"" + command + "\"");
#endif
#endif
  qApp->processEvents();
#ifdef SDLMAME
  commandProc.start(qmc2Config->value("MAME/FilesAndDirectories/ExecutableFile").toString(), args);
  commandProc.waitForFinished(-1);
#else
  int exitCode = QProcess::execute(command);
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: command finished: exitCode = " + QString::number(exitCode));
#endif
#endif
  QFile qmc2Temp(qmc2Options->config->value("Frontend/FilesAndDirectories/TemporaryFile").toString());
  if ( qmc2Temp.open(QFile::ReadOnly) ) {
    QTextStream ts(&qmc2Temp);
#ifdef SDLMAME
    QString s = ts.readAll();
#else
    QString s = ts.readAll().simplified();
#endif
    qmc2Temp.close();
    qmc2Temp.remove();
#ifdef SDLMAME
    numTotalGames = s.split("\n").count() - 2;
#else
    s = s.remove(0, s.lastIndexOf("Total Supported: ")).remove("Total Supported: ");
    numTotalGames = s.toInt();
#endif
    qmc2MainWindow->log(LOG_QMC2, QString::number(numTotalGames) + tr(" games total"));
  } else {
    qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't create temporary file"));
  }
  qmc2MainWindow->labelGamelistStatus->setText(status());

  if ( numTotalGames <= 0 ) {
    qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't determine total number of supported games"));
    qmc2ReloadActive = FALSE;
    enableWidgets(TRUE);
    return;
  }

  loadTimer.start();
  qmc2MainWindow->log(LOG_QMC2, tr("loading gamelist data"));
  
  command = qmc2Config->value("MAME/FilesAndDirectories/ExecutableFile").toString();

  args.clear();
#ifdef SDLMAME
  args << "-listxml";
#else
  args << "-lx";
#endif

  gamelistBuffer = "";
  loadProc = qmc2ProcessManager->process(qmc2ProcessManager->start(command, args, FALSE));
  connect(loadProc, SIGNAL(error(QProcess::ProcessError)), this, SLOT(loadError(QProcess::ProcessError)));
  connect(loadProc, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(loadFinished(int, QProcess::ExitStatus)));
  connect(loadProc, SIGNAL(readyReadStandardOutput()), this, SLOT(loadReadyReadStandardOutput()));
  connect(loadProc, SIGNAL(readyReadStandardError()), this, SLOT(loadReadyReadStandardError()));
  connect(loadProc, SIGNAL(started()), this, SLOT(loadStarted()));
  connect(loadProc, SIGNAL(stateChanged(QProcess::ProcessState)), this, SLOT(loadStateChanged(QProcess::ProcessState)));
}

void Gamelist::verify()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verify()");
#endif

  qmc2VerifyActive = TRUE;
  qmc2StopParser = FALSE;

  enableWidgets(FALSE);

#ifdef SDLMAME
  verifiedList.clear();
#endif

  verifyLastLine = "";
  verifyTimer.start();
  numVerifyRoms = 0;
  qmc2MainWindow->log(LOG_QMC2, tr("verifying ROM status"));
  
  QString command = qmc2Config->value("MAME/FilesAndDirectories/ExecutableFile").toString();
  QStringList args;

  args << "-verifyroms";
  args << "-rompath" << qmc2Config->value("MAME/Configuration/Global/rompath").toString();

  verifyProc = qmc2ProcessManager->process(qmc2ProcessManager->start(command, args, FALSE));
  connect(verifyProc, SIGNAL(error(QProcess::ProcessError)), this, SLOT(verifyError(QProcess::ProcessError)));
  connect(verifyProc, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(verifyFinished(int, QProcess::ExitStatus)));
  connect(verifyProc, SIGNAL(readyReadStandardOutput()), this, SLOT(verifyReadyReadStandardOutput()));
  connect(verifyProc, SIGNAL(readyReadStandardError()), this, SLOT(verifyReadyReadStandardError()));
  connect(verifyProc, SIGNAL(started()), this, SLOT(verifyStarted()));
  connect(verifyProc, SIGNAL(stateChanged(QProcess::ProcessState)), this, SLOT(verifyStateChanged(QProcess::ProcessState)));
}

QString Gamelist::value(QString element, QString attribute, bool translate)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::value(QString element = " + element + ", QString attribute = \"" + attribute + "\", translate = " + QString(translate ? "TRUE" : "FALSE") + ")");
#endif

  if ( element.contains(attribute) ) {
    QString valueString = element.remove(0, element.indexOf(attribute) + attribute.length() + 2);
    valueString = valueString.remove(valueString.indexOf("\""), valueString.lastIndexOf(">")).replace("&amp;", "&");
    if ( translate )
      return tr(valueString.toAscii());
    else
      return valueString;
  }
  else {
    return QString::null;
  }
}

void Gamelist::insertAttributeItems(QTreeWidgetItem *parent, QString element, QStringList attributes, QStringList descriptions, bool translate)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::insertAttributeItems(QTreeWidgetItem *parent = 0x" + QString::number((ulong)parent, 16) + ", QString element = \"" + element + "\", QStringList attributes = ..., QStringList descriptions = ..., translate = " + QString(translate ? "TRUE" : "FALSE") + ")");
#endif

  int i;
  for (i = 0; i < attributes.count(); i++) {
    QString valueString = value(element, attributes.at(i), translate);
    if ( !valueString.isEmpty() ) {
      QTreeWidgetItem *attributeItem = new QTreeWidgetItem(parent);
      attributeItem->setText(0, descriptions.at(i));
      attributeItem->setText(1, tr(valueString.toAscii()));
    }
  }
}

void Gamelist::parseGameDetail(QTreeWidgetItem *item)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::parseGameDetail(QTreeWidgetItem *item = 0x" + QString::number((ulong)item, 16) + "): item->text(0) = \"" + item->text(0) + "\"");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("retrieving game information for ") + item->text(0));
  qApp->processEvents();

  int gamePos = 0;
  QString s = "<description>" + item->text(0).replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;") + "</description>";
  while ( !xmlLines[gamePos].contains(s) ) gamePos++;

  QTreeWidgetItem *childItem = item->takeChild(0);
  delete childItem;

  QString element, content;
  QStringList attributes, descriptions;

  // game element
  attributes << "name" << "sourcefile" << "runnable" << "cloneof" << "romof" << "sampleof" ;
  descriptions << tr("Name") << tr("Source file") << tr("Runnable") << tr("Clone of") << tr("ROM of") << tr("Sample of");
  element = xmlLines.at(gamePos - 1).simplified();
  insertAttributeItems(item, element, attributes, descriptions, TRUE);

  while ( !xmlLines[gamePos].contains("</game>") ) {
    element = xmlLines[gamePos].simplified();
    if ( element.contains("<year>") ) {
      content = element.remove("<year>").remove("</year>");
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Year"));
      childItem->setText(1, content);
    }
    if ( element.contains("<manufacturer>") ) {
      content = element.remove("<manufacturer>").remove("</manufacturer>");
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Manufacturer"));
      childItem->setText(1, content);
    }
    if ( element.contains("<rom ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("ROM"));
      childItem->setText(1, value(element, "name"));
      attributes.clear();
      attributes << "bios" << "size" << "crc" << "sha1" << "merge" << "region" << "offset" << "status" << "dispose";
      descriptions.clear();
      descriptions << tr("BIOS") << tr("Size") << tr("CRC") << tr("SHA1") << tr("Merge") << tr("Region") << tr("Offset") << tr("Status") << tr("Dispose");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<chip ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Chip"));
      childItem->setText(1, value(element, "name"));
      attributes.clear();
      attributes << "type" << "clock";
      descriptions.clear();
      descriptions << tr("Type") << tr("Clock");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<video ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Video"));
      attributes.clear();
      attributes << "screen" << "orientation" << "width" << "height" << "aspectx" << "aspecty" << "refresh";
      descriptions.clear();
      descriptions << tr("Screen") << tr("Orientation") << tr("Width") << tr("Height") << tr("Aspect X") << tr("Aspect Y") << tr("Refresh");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<sound ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Sound"));
      attributes.clear();
      attributes << "channels";
      descriptions.clear();
      descriptions << tr("Channels");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<input ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Input"));
      attributes.clear();
      attributes << "service" << "tilt" << "players" << "control" << "buttons" << "coins";
      descriptions.clear();
      descriptions << tr("Service") << tr("Tilt") << tr("Players") << tr("Control") << tr("Buttons") << tr("Coins");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<dipswitch ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("DIP switch"));
      childItem->setText(1, value(element, "name"));
      gamePos++;
      while ( xmlLines[gamePos].contains("<dipvalue ") ) {
        QString subElement = xmlLines[gamePos].simplified();
        QTreeWidgetItem *secondChildItem = new QTreeWidgetItem(childItem);
        secondChildItem->setText(0, tr("DIP value"));
        secondChildItem->setText(1, value(subElement, "name"));
        attributes.clear();
        attributes << "default";
        descriptions.clear();
        descriptions << tr("Default");
        insertAttributeItems(secondChildItem, subElement, attributes, descriptions, TRUE);
        gamePos++;
      }
    }
    if ( element.contains("<driver ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Driver"));
      attributes.clear();
      attributes << "status" << "emulation" << "color" << "sound" << "graphic" << "cocktail" << "protection" << "savestate" << "palettesize";
      descriptions.clear();
      descriptions << tr("Status") << tr("Emulation") << tr("Color") << tr("Sound") << tr("Graphic") << tr("Cocktail") << tr("Protection") << tr("Save state") << tr("Palette size");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<biosset ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("BIOS set"));
      childItem->setText(1, value(element, "name"));
      attributes.clear();
      attributes << "description" << "default";
      descriptions.clear();
      descriptions << tr("Description") << tr("Default");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    if ( element.contains("<sample ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Sample"));
      childItem->setText(1, value(element, "name"));
    }
    if ( element.contains("<disk ") ) {
      childItem = new QTreeWidgetItem(item);
      childItem->setText(0, tr("Disk"));
      childItem->setText(1, value(element, "name"));
      attributes.clear();
      attributes << "md5" << "sha1" << "merge" << "region" << "index" << "status";
      descriptions.clear();
      descriptions << tr("MD5") << tr("SHA1") << tr("Merge") << tr("Region") << tr("Index") << tr("Status");
      insertAttributeItems(childItem, element, attributes, descriptions, TRUE);
    }
    gamePos++;
  }
}

void Gamelist::parse()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::parse()");
#endif

  if ( qmc2StopParser ) {
    qmc2ReloadActive = FALSE;
    enableWidgets(TRUE);
    return;
  }

  QTime elapsedTime;
  qmc2MainWindow->progressBarGamelist->setRange(0, numTotalGames);
  romCache.setFileName(qmc2Config->value("Frontend/FilesAndDirectories/ROMStateCacheFile").toString());
  romCache.open(QIODevice::ReadOnly | QIODevice::Text);
  if ( !romCache.isOpen() ) {
    qmc2MainWindow->log(LOG_QMC2, tr("WARNING: can't open ROM state cache, please check ROMs"));
  } else {
    parseTimer.start();
    qmc2MainWindow->log(LOG_QMC2, tr("loading ROM state from cache"));
    qmc2MainWindow->progressBarGamelist->reset();
    tsRomCache.setDevice(&romCache);
    tsRomCache.reset();
    int cachedGamesCounter = 0;
    while ( !tsRomCache.atEnd() ) {
      QString line = tsRomCache.readLine();
      if ( !line.isNull() ) {
        QStringList words = line.split(" ");
        qmc2GamelistStatusMap[words[0]] = words[1];
        cachedGamesCounter++;
        qmc2MainWindow->progressBarGamelist->setValue(cachedGamesCounter);
      }
      qApp->processEvents();
    }
    numCorrectGames = numMostlyCorrectGames = numIncorrectGames = numNotFoundGames = 0;
    elapsedTime = elapsedTime.addMSecs(parseTimer.elapsed());
    qmc2MainWindow->log(LOG_QMC2, tr("done (loading ROM state from cache, elapsed time = %1) - %2 cached game states loaded").arg(elapsedTime.toString("mm:ss.zzz")).arg(cachedGamesCounter));
    if ( numTotalGames != cachedGamesCounter )
      qmc2MainWindow->log(LOG_QMC2, tr("WARNING: ROM state cache is incomplete or not up to date, please re-check ROMs"));
  }
  if ( romCache.isOpen() )
    romCache.close();

  qmc2MainWindow->log(LOG_QMC2, tr("parsing gamelist data"));
  parseTimer.start();
  qmc2MainWindow->treeWidgetGamelist->clear();
  qmc2MainWindow->progressBarGamelist->reset();

  // parse XML gamelist data
  xmlLines.clear();
  xmlLines = gamelistBuffer.remove(0, gamelistBuffer.indexOf("<mame build")).split('\n');
  gamelistBuffer.clear();

  int lineCounter;
  numGames = numUnknownGames = 0;
  bool endParser = qmc2StopParser;
  qmc2MainWindow->treeWidgetGamelist->setUpdatesEnabled(FALSE);

  for (lineCounter = 0; xmlLines.count() && !endParser; lineCounter++) {
    while ( !endParser && !xmlLines[lineCounter].contains("<description>") ) {
      lineCounter++;
      endParser = xmlLines[lineCounter].contains("</mame>") || qmc2StopParser;
    }
    if ( !endParser ) {
      QString descriptionElement = xmlLines[lineCounter].simplified();
      QString gameElement = xmlLines[lineCounter - 1].simplified();
      QString runnableCheck = value(gameElement, "runnable");
      if ( runnableCheck != "no" ) {
        QString gameName = value(gameElement, "name");
        QString gameDescription = descriptionElement.remove("<description>").remove("</description>").replace("&amp;", "&").replace("&lt;", "<").replace("&gt;", ">");
        GamelistItem *gameDescriptionItem = new GamelistItem(qmc2MainWindow->treeWidgetGamelist);

        gameDescriptionItem->setText(0, gameDescription);
        switch ( qmc2GamelistStatusMap[gameName][0].toAscii() ) {
          case 'C': 
            numCorrectGames++;
            gameDescriptionItem->setIcon(0, qmc2CorrectImageIcon);
            gameDescriptionItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_C);
            break;

          case 'M': 
            numMostlyCorrectGames++;
            gameDescriptionItem->setIcon(0, qmc2MostlyCorrectImageIcon);
            gameDescriptionItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_M);
            break;

          case 'I':
            numIncorrectGames++;
            gameDescriptionItem->setIcon(0, qmc2IncorrectImageIcon);
            gameDescriptionItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_I);
            break;

          case 'N':
            numNotFoundGames++;
            gameDescriptionItem->setIcon(0, qmc2NotFoundImageIcon);
            gameDescriptionItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_N);
            break;

          default:
            numUnknownGames++;
            qmc2GamelistStatusMap[gameName] = "U";
            gameDescriptionItem->setIcon(0, qmc2UnknownImageIcon);
            gameDescriptionItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_U);
            break;
        }

        QTreeWidgetItem *nameItem = new QTreeWidgetItem(gameDescriptionItem);
        nameItem->setText(0, tr("Waiting for data..."));
        nameItem->setText(1, gameName);
        qmc2GamelistItemMap[gameName] = gameDescriptionItem;
        qmc2GamelistNameMap[gameDescription] = gameName;

        numGames++;
      }
    }

    if ( numGames % qmc2GamelistResponsiveness == 0 ) {
      qmc2MainWindow->treeWidgetGamelist->setUpdatesEnabled(TRUE);
      qmc2MainWindow->labelGamelistStatus->setText(status());
      if ( qmc2Options->config->value("Frontend/Gamelist/SortOnline").toBool() )
        qmc2MainWindow->treeWidgetGamelist->sortItems(0, qmc2SortOrder);
      qApp->processEvents();
      qmc2MainWindow->treeWidgetGamelist->setUpdatesEnabled(FALSE);
    }
    qApp->processEvents();

    qmc2MainWindow->progressBarGamelist->setValue(numGames);
  }

  qmc2MainWindow->log(LOG_QMC2, tr("sorting gamelist by %1 in %2 order").arg(qmc2SortCriteria == QMC2_SORT_BY_DESCRIPTION ? tr("game description") : tr("ROM state")).arg(qmc2SortOrder == Qt::AscendingOrder ? tr("ascending") : tr("descending")));
  qmc2MainWindow->treeWidgetGamelist->sortItems(0, qmc2SortOrder);
  QTreeWidgetItem *ci = qmc2MainWindow->treeWidgetGamelist->currentItem();
  if ( ci ) {
    if ( ci->isSelected() ) {
      QTimer::singleShot(0, qmc2MainWindow, SLOT(scrollToCurrentItem()));
    } else if ( qmc2Config->value("Frontend/GUI/RestoreGameSelection").toBool() ) {
      QList<QTreeWidgetItem *> gl = qmc2MainWindow->treeWidgetGamelist->findItems(qmc2Config->value("Frontend/Gamelist/SelectedGame").toString(), Qt::MatchExactly, 0);
      if ( gl.count() > 0 ) {
        qmc2MainWindow->log(LOG_QMC2, tr("restoring game selection"));
        qmc2MainWindow->treeWidgetGamelist->setCurrentItem(gl.at(0));
        QTimer::singleShot(0, qmc2MainWindow, SLOT(scrollToCurrentItem()));
      }
    }
  } else if ( qmc2Config->value("Frontend/GUI/RestoreGameSelection").toBool() ) {
    QList<QTreeWidgetItem *> gl = qmc2MainWindow->treeWidgetGamelist->findItems(qmc2Config->value("Frontend/Gamelist/SelectedGame").toString(), Qt::MatchExactly, 0);
    if ( gl.count() > 0 ) {
      qmc2MainWindow->log(LOG_QMC2, tr("restoring game selection"));
      qmc2MainWindow->treeWidgetGamelist->setCurrentItem(gl.at(0));
      QTimer::singleShot(0, qmc2MainWindow, SLOT(scrollToCurrentItem()));
    }
  }

  qmc2MainWindow->treeWidgetGamelist->setUpdatesEnabled(TRUE);
  qmc2MainWindow->labelGamelistStatus->setText(status());

  elapsedTime = elapsedTime.addMSecs(parseTimer.elapsed());
  qmc2MainWindow->log(LOG_QMC2, tr("done (parsing gamelist data, elapsed time = %1)").arg(elapsedTime.toString("mm:ss.zzz")));
  qmc2MainWindow->progressBarGamelist->reset();

  if ( qmc2StopParser && loadProc )
    loadProc->terminate();

  qmc2ReloadActive = FALSE;
  qmc2StartingUp = FALSE;

  enableWidgets(TRUE);

  filter();
}

void Gamelist::filter()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::filter()");
#endif

  if ( qmc2FilterActive ) {
    qmc2MainWindow->log(LOG_QMC2, tr("ROM state filter already active"));
    return;
  }

  int i;
  QTime elapsedTime;
  parseTimer.start();
  qmc2FilterActive = TRUE;
  enableWidgets(FALSE);
  qmc2MainWindow->log(LOG_QMC2, tr("applying ROM state filter"));
  qmc2MainWindow->progressBarGamelist->reset();
  qmc2MainWindow->progressBarGamelist->setRange(0, numGames);
  bool showC = qmc2Config->value("Frontend/Gamelist/ShowC").toBool();
  bool showM = qmc2Config->value("Frontend/Gamelist/ShowM").toBool();
  bool showI = qmc2Config->value("Frontend/Gamelist/ShowI").toBool();
  bool showN = qmc2Config->value("Frontend/Gamelist/ShowN").toBool();
  bool showU = qmc2Config->value("Frontend/Gamelist/ShowU").toBool();
  for (i = 0; i < numGames && !qmc2StopParser; i++) {
    QTreeWidgetItem *item = qmc2MainWindow->treeWidgetGamelist->topLevelItem(i);
    switch ( item->whatsThis(0).at(0).toAscii() ) {
      case QMC2_ROMSTATE_CHAR_C:
        item->setHidden(!showC);
        break;

      case QMC2_ROMSTATE_CHAR_M:
        item->setHidden(!showM);
        break;

      case QMC2_ROMSTATE_CHAR_I:
        item->setHidden(!showI);
        break;

      case QMC2_ROMSTATE_CHAR_N:
        item->setHidden(!showN);
        break;

      case QMC2_ROMSTATE_CHAR_U:
        item->setHidden(!showU);
        break;
    }
    qmc2MainWindow->progressBarGamelist->setValue(i);
    if ( i % QMC2_STATEFILTER_RESPONSIVENESS == 0 ) {
      qApp->processEvents();
    }
  }
  qmc2MainWindow->progressBarGamelist->reset();
  enableWidgets(TRUE);
  qmc2FilterActive = FALSE;
  elapsedTime = elapsedTime.addMSecs(parseTimer.elapsed());
  qmc2MainWindow->log(LOG_QMC2, tr("done (applying ROM state filter, elapsed time = %1)").arg(elapsedTime.toString("mm:ss.zzz")));
}

void Gamelist::save()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::save()");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("saving gamelist"));

  qmc2MainWindow->log(LOG_QMC2, tr("done (saving gamelist)"));
}

void Gamelist::loadFavorites()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadFavorites()");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("loading favorites"));

  qmc2MainWindow->listWidgetFavorites->clear();
  QFile f(qmc2Config->value("Frontend/FilesAndDirectories/FavoritesFile").toString());
  if ( f.open(QIODevice::ReadOnly | QIODevice::Text) ) {
    QTextStream ts(&f);
    while ( !ts.atEnd() ) {
      QString gameName = ts.readLine();
      if ( !gameName.isEmpty() ) {
        QTreeWidgetItem *gameItem = qmc2GamelistItemMap[gameName];
        if ( gameItem ) {
          QListWidgetItem *item = new QListWidgetItem(qmc2MainWindow->listWidgetFavorites);
          item->setText(gameItem->text(0));
        }
      }
    }
    f.close();
  }

  qmc2MainWindow->listWidgetFavorites->sortItems();
  qmc2MainWindow->log(LOG_QMC2, tr("done (loading favorites)"));
  QTimer::singleShot(0, qmc2MainWindow, SLOT(checkCurrentFavoritesSelection()));
}

void Gamelist::saveFavorites()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::saveFavorites()");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("saving favorites"));

  QFile f(qmc2Config->value("Frontend/FilesAndDirectories/FavoritesFile").toString());
  if ( f.open(QIODevice::WriteOnly | QIODevice::Text) ) {
    QTextStream ts(&f);
    int i;
    for (i = 0; i < qmc2MainWindow->listWidgetFavorites->count(); i++) {
      ts << qmc2GamelistNameMap[qmc2MainWindow->listWidgetFavorites->item(i)->text()] << "\n";
    }
    f.close();
  } else {
    qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open favorites file for writing, path = %1").arg(qmc2Config->value("Frontend/FilesAndDirectories/FavoritesFile").toString()));
  }

  qmc2MainWindow->log(LOG_QMC2, tr("done (saving favorites)"));
}

void Gamelist::loadPlayHistory()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadPlayHistory()");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("loading play history"));

  qmc2MainWindow->listWidgetPlayed->clear();
  QFile f(qmc2Config->value("Frontend/FilesAndDirectories/HistoryFile").toString());
  if ( f.open(QIODevice::ReadOnly | QIODevice::Text) ) {
    QTextStream ts(&f);
    while ( !ts.atEnd() ) {
      QString gameName = ts.readLine();
      if ( !gameName.isEmpty() ) {
        QTreeWidgetItem *gameItem = qmc2GamelistItemMap[gameName];
        if ( gameItem ) {
          QListWidgetItem *item = new QListWidgetItem(qmc2MainWindow->listWidgetPlayed);
          item->setText(gameItem->text(0));
        }
      }
    }
    f.close();
  }

  qmc2MainWindow->log(LOG_QMC2, tr("done (loading play history)"));
  QTimer::singleShot(0, qmc2MainWindow, SLOT(checkCurrentPlayedSelection()));
}

void Gamelist::savePlayHistory()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::savePlayHistory()");
#endif

  qmc2MainWindow->log(LOG_QMC2, tr("saving play history"));

  QFile f(qmc2Config->value("Frontend/FilesAndDirectories/HistoryFile").toString());
  if ( f.open(QIODevice::WriteOnly | QIODevice::Text) ) {
    QTextStream ts(&f);
    int i;
    for (i = 0; i < qmc2MainWindow->listWidgetPlayed->count(); i++) {
      ts << qmc2GamelistNameMap[qmc2MainWindow->listWidgetPlayed->item(i)->text()] << "\n";
    }
    f.close();
  } else {
    qmc2MainWindow->log(LOG_QMC2, tr("FATAL: can't open play history file for writing, path = %1").arg(qmc2Config->value("Frontend/FilesAndDirectories/HistoryFile").toString()));
  }

  qmc2MainWindow->log(LOG_QMC2, tr("done (saving play history)"));
}

QString Gamelist::status()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::status()");
#endif

  QString statusString = "<b>";
  statusString += "<font color=black>" + tr("L:") + QString(numGames > -1 ? QString::number(numGames) : "?") + "</font> ";
  statusString += "<font color=green>" + tr("C:") + QString(numCorrectGames > -1 ? QString::number(numCorrectGames) : "?") + "</font> ";
  statusString += "<font color=yellowgreen>" + tr("M:") + QString(numMostlyCorrectGames > -1 ? QString::number(numMostlyCorrectGames) : "?") + "</font> ";
  statusString += "<font color=red>" + tr("I:") + QString(numIncorrectGames > -1 ? QString::number(numIncorrectGames) : "?") + "</font> ";
  statusString += "<font color=grey>" + tr("N:") + QString(numNotFoundGames > -1 ? QString::number(numNotFoundGames) : "?") + "</font> ";
  statusString += "<font color=blue>" + tr("U:") + QString(numUnknownGames > -1 ? QString::number(numUnknownGames) : "?") + "</font> ";
  statusString += "<font color=chocolate>" + tr("S:") + QString(numSearchGames > -1 ? QString::number(numSearchGames) : "?") + "</font>";
  statusString += "</b>";

  return statusString;
}

void Gamelist::loadStarted()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadStarted()");
#endif

  qmc2MainWindow->progressBarGamelist->setRange(0, numTotalGames);
  qmc2MainWindow->progressBarGamelist->reset();
}

void Gamelist::loadFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
  QProcess *proc = (QProcess *)sender();

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadFinished(int exitCode = " + QString::number(exitCode) + ", QProcess::ExitStatus exitStatus = " + QString::number(exitStatus) + "): proc = 0x" + QString::number((ulong)proc, 16));
#endif

  QTime elapsedTime;
  elapsedTime = elapsedTime.addMSecs(loadTimer.elapsed());
  qmc2MainWindow->log(LOG_QMC2, tr("done (loading gamelist data, elapsed time = %1)").arg(elapsedTime.toString("mm:ss.zzz")));
  qmc2MainWindow->progressBarGamelist->reset();
  qmc2ProcessManager->procMap.remove(proc);
  qmc2ProcessManager->procCount--;
  qmc2EarlyReloadActive = FALSE;
  loadProc = NULL;

  if ( romCache.isOpen() )
    romCache.close();

  parse();

  loadFavorites();
  loadPlayHistory();
}

void Gamelist::loadReadyReadStandardOutput()
{
  QProcess *proc = (QProcess *)sender();

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadReadyReadStandardOutput(): proc = 0x" + QString::number((ulong)proc, 16));
#endif

  QString s = proc->readAllStandardOutput();

  qmc2MainWindow->progressBarGamelist->setValue(qmc2MainWindow->progressBarGamelist->value() + s.count("<game name="));
  gamelistBuffer += s;
}

void Gamelist::loadReadyReadStandardError()
{
  QProcess *proc = (QProcess *)sender();

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadReadyReadStandardError(): proc = 0x" + QString::number((ulong)proc, 16));
#endif

}

void Gamelist::loadError(QProcess::ProcessError processError)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadError(QProcess::ProcessError processError = " + QString::number(processError) + ")");
#endif

}

void Gamelist::loadStateChanged(QProcess::ProcessState processState)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::loadStateChanged(QProcess::ProcessState = " + QString::number(processState) + ")");
#endif

}

void Gamelist::verifyStarted()
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyStarted()");
#endif

  romCache.setFileName(qmc2Config->value("Frontend/FilesAndDirectories/ROMStateCacheFile").toString());
  romCache.open(QIODevice::WriteOnly | QIODevice::Text);
  if ( !romCache.isOpen() )
    qmc2MainWindow->log(LOG_QMC2, tr("ERROR: can't open ROM verification cache for writing, path = %1").arg(romCache.fileName()));
  else {
    tsRomCache.setDevice(&romCache);
    tsRomCache.reset();
  }
  qmc2MainWindow->progressBarGamelist->setRange(0, numTotalGames);
  qmc2MainWindow->progressBarGamelist->reset();
  numCorrectGames = numMostlyCorrectGames = numIncorrectGames = numNotFoundGames = numUnknownGames = 0;
}

void Gamelist::verifyFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
  QProcess *proc = (QProcess *)sender();

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyFinished(int exitCode = " + QString::number(exitCode) + ", QProcess::ExitStatus exitStatus = " + QString::number(exitStatus) + "): proc = 0x" + QString::number((ulong)proc, 16));
#endif

#ifdef SDLMAME
  QSet<QString> gameSet = QSet<QString>::fromList(qmc2GamelistItemMap.uniqueKeys());
  QList<QString> remainingGames = gameSet.subtract(QSet<QString>::fromList(verifiedList)).values();
  int i;
  if ( qmc2StopParser ) {
    for (i = 0; i < remainingGames.count(); i++) {
      if ( romCache.isOpen() )
          tsRomCache << remainingGames[i] << " U\n";
      numUnknownGames++;
      qmc2GamelistItemMap[remainingGames[i]]->setIcon(0, qmc2UnknownImageIcon);
      qmc2GamelistItemMap[remainingGames[i]]->setWhatsThis(0, QMC2_ROMSTATE_STRING_U);
    }
  } else {
    for (i = 0; i < remainingGames.count(); i++) {
      if ( romCache.isOpen() )
          tsRomCache << remainingGames[i] << " N\n";
      numNotFoundGames++;
      qmc2GamelistItemMap[remainingGames[i]]->setIcon(0, qmc2NotFoundImageIcon);
      qmc2GamelistItemMap[remainingGames[i]]->setWhatsThis(0, QMC2_ROMSTATE_STRING_N);
    }
  }
  qmc2MainWindow->labelGamelistStatus->setText(status());
#endif

  QTime elapsedTime;
  elapsedTime = elapsedTime.addMSecs(verifyTimer.elapsed());
  qmc2MainWindow->log(LOG_QMC2, tr("done (verifying ROMs, elapsed time = %1)").arg(elapsedTime.toString("mm:ss.zzz")));
  qmc2MainWindow->progressBarGamelist->reset();
  qmc2ProcessManager->procMap.remove(proc);
  qmc2ProcessManager->procCount--;
  qmc2VerifyActive = FALSE;
  verifyProc = NULL;
  if ( romCache.isOpen() ) {
    tsRomCache.flush();
    romCache.close();
  }

  if ( qmc2SortCriteria == QMC2_SORT_BY_ROM_STATE ) {
    qmc2MainWindow->log(LOG_QMC2, tr("sorting gamelist by %1 in %2 order").arg(tr("ROM state")).arg(qmc2SortOrder == Qt::AscendingOrder ? tr("ascending") : tr("descending")));
    qmc2MainWindow->treeWidgetGamelist->sortItems(0, qmc2SortOrder);
    QTreeWidgetItem *ci = qmc2MainWindow->treeWidgetGamelist->currentItem();
    if ( ci )
      QTimer::singleShot(0, qmc2MainWindow, SLOT(scrollToCurrentItem()));
  }

  enableWidgets(TRUE);

  filter();
}

void Gamelist::verifyReadyReadStandardOutput()
{
  QProcess *proc = (QProcess *)sender();

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyReadyReadStandardOutput(): proc = 0x" + QString::number((ulong)proc, 16));
#endif

  // process rom verification output
  int i;
  QString romName, romState; 
  QString s = verifyLastLine + proc->readAllStandardOutput();
  QStringList lines = s.split("\n");

  if ( s.endsWith("\n") ) {
    verifyLastLine = "";
  } else {
    verifyLastLine = lines.last();
    lines.removeLast();
  }

  for (i = 0; i < lines.count(); i++) {
    if ( lines[i].startsWith("romset ") ) {
      QStringList words = lines[i].split(" ");
      numVerifyRoms++;
      if ( words.count() > 2 ) {
        romName = words[1];
        if ( qmc2GamelistItemMap.count(romName) == 1 ) {
#ifdef SDLMAME
          QTreeWidgetItem *romItem = qmc2GamelistItemMap[romName];
          if ( words.last() == "good" ) {
            romState = "C";
            numCorrectGames++;
            romItem->setIcon(0, qmc2CorrectImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_C);
          } else if ( words.last() == "bad" ) {
            romState = "I";
            numIncorrectGames++;
            romItem->setIcon(0, qmc2IncorrectImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_I);
          } else if ( words.last() == "available" ) {
            romState = "M";
            numMostlyCorrectGames++;
            romItem->setIcon(0, qmc2MostlyCorrectImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_M);
          } else if ( words.last() == "missing" ) {
            romState = "N";
            numNotFoundGames++;
            romItem->setIcon(0, qmc2NotFoundImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_N);
          } else {
            romState = "U";
            numUnknownGames++;
            romItem->setIcon(0, qmc2UnknownImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_U);
          }
#else
          if ( words[2] == "correct" ) {
            romState = "C";
            numCorrectGames++;
            romItem->setIcon(0, qmc2CorrectImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_C);
          } else if ( words[2] == "incorrect" ) {
            romState = "I";
            numIncorrectGames++;
            romItem->setIcon(0, qmc2IncorrectImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_I);
          } else if ( words[2] == "best" ) {
            romState = "M";
            numMostlyCorrectGames++;
            romItem->setIcon(0, qmc2MostlyCorrectImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_M);
          } else if ( words[2] == "not" ) {
            romState = "N";
            numNotFoundGames++;
            romItem->setIcon(0, qmc2NotFoundImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_N);
          } else {
            romState = "U";
            numUnknownGames++;
            romItem->setIcon(0, qmc2UnknownImageIcon);
            romItem->setWhatsThis(0, QMC2_ROMSTATE_STRING_U);
          }
#endif

#ifdef QMC2_DEBUG
          qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyReadyReadStandardOutput(): " + romName + " " + romState);
#endif

          qmc2GamelistStatusMap[romName] = romState;

#ifdef SDLMAME
          verifiedList << romName;
#endif

          if ( romCache.isOpen() )
            tsRomCache << romName << " " << romState << "\n";
        }
      }
    }
  }
  if ( romCache.isOpen() )
    tsRomCache.flush();

  if ( qmc2StopParser && verifyProc )
    verifyProc->terminate();

  qmc2MainWindow->progressBarGamelist->setValue(numVerifyRoms);
  qmc2MainWindow->labelGamelistStatus->setText(status());
}

void Gamelist::verifyReadyReadStandardError()
{
  QProcess *proc = (QProcess *)sender();

#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyReadyReadStandardError(): proc = 0x" + QString::number((ulong)proc, 16));
#endif

}

void Gamelist::verifyError(QProcess::ProcessError processError)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyError(QProcess::ProcessError processError = " + QString::number(processError) + ")");
#endif

}

void Gamelist::verifyStateChanged(QProcess::ProcessState processState)
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: Gamelist::verifyStateChanged(QProcess::ProcessState = " + QString::number(processState) + ")");
#endif

}

bool GamelistItem::operator<(const QTreeWidgetItem &otherItem) const
{
#ifdef QMC2_DEBUG
  qmc2MainWindow->log(LOG_QMC2, "DEBUG: GamelistItem::operator<(const GamelistItem &otherItem = ...)");
#endif
  
  switch ( qmc2SortCriteria ) {
    case QMC2_SORT_BY_DESCRIPTION:
      return (text(0).toUpper() < otherItem.text(0).toUpper());
      break;

    case QMC2_SORT_BY_ROM_STATE:
      return (whatsThis(0) < otherItem.whatsThis(0));
      break;

    default:
      return FALSE;
      break;
  }
}
